// Comment

#include <stdio.h>

int main (void)
{
    int a = 0x89;
    int b = 089;
    int c = 89.;
    int d = 'a';
    double hexadecimal_floating_constant = 0x1.2p3;
    printf ("Hello %s!\n", "world");
    return 0;
}
